"""
Child class of ArduinoComBase that handles the communication with the target
Arduino.
"""

import binascii
from .arduino_com_base import ArduinoComBase
from .arduino_com_result import Result, Ok, Err

class ArduinoFormationFaults(ArduinoComBase):
    """
    Class that handles communication with the Arduino board on which errors
    are injected.
    """

    def set_led_on(self) -> Result[None]:
        self._safe_instrument_reset_input_buffer()
        match self._safe_write_then_readline("H"):
            case Ok("OK\r\n"):
                return Ok(None)
            case Ok(answer):
                return Err(Exception(
                    f"Unexpected response from Arduino: {answer}."
                ))
            case err:
                return err

    def set_led_off(self) -> Result[None]:
        self._safe_instrument_reset_input_buffer()
        match self._safe_write_then_readline("L"):
            case Ok("OK\r\n"):
                return Ok(None)
            case Ok(answer):
                return Err(Exception(
                    f"Unexpected response from Arduino: {answer}."
                ))
            case err:
                return err

    def set_asmcode(self) -> Result[None]:
        """
        Start a series of 10 LDI assembler instructions.
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("A"):
                    case Ok("OK\r\n"):
                        return Ok(None)
                    case Ok(answer):
                        return Err(Exception(
                            f"Received something else than OK: {answer}"
                        ))
                    case err:
                        return err
            case err:
                return err

    def get_regs(self) -> Result[list[int]]:
        """
        Read the 10 registers R16 to R25 from the SRAM.

        Returns something like if the communication worked :
        Ok([0x39, 0x38, 0x37, 0x36, 0x35, 0x34, 0x33, ..., 0x30]).
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("Z", 0.2):
                    case Ok(regs_str):
                        # received `regs_str` format should be :
                        # "0x39 0x38 0x37 0x36 0x35 0x34 0x33 0x32 0x31 0x30"
                        regs_str = regs_str.rstrip()
                        try:
                            regs = list(
                                map(lambda x: int(x, 0), regs_str.split(" "))
                            )
                        except ValueError:
                            return Err(Exception(
                                f"Could not parse received regs: {regs_str}"
                            ))

                        if len(regs) != 10:
                            return Err(Exception(
                                f"Received less/more than 10 regs: {regs}"
                            ))
                        else:
                            return Ok(regs)
                    case err:
                        return err
            case err:
                return err

    def set_user_pin(self, pin : int) -> Result[None]:
        """
        Set the PIN code entered by the user.

        Args:
            - pin (int): the 4 digits PIN code.

        """
        # clamp pin to 4 digits
        pin = min(9999, max(0, pin))

        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline(
                    "P" + str(pin).zfill(4),
                    sleep_before_read_s=0.1
                ):
                    case Ok("OK\r\n"):
                        return Ok(None)
                    case Ok(answer):
                        return Err(Exception(
                            f"Received something else than OK: {answer}"
                        ))
                    case err:
                        return err
            case err:
                return err

    def get_user_pin(self) -> Result[int]:
        """
        Get the PIN code entered by the user [0;9999].
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_read("R", nb_bytes=4):
                    case Ok(pin_str):
                        return Ok(min(9999, max(0, int(pin_str))))
                    case err:
                        return err
            case err:
                return err

    def verifypin(self) -> Result[bool]:
        """
        Verify the PIN code entered by the user.

        Returns:
            - response (Result): answer from Arduino (True(0xAA) (successful
              attack) or False(0x55) (no double fault)).
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("V"):
                    case Ok(status_str):
                        status_str = status_str.rstrip()
                        try:
                            return Ok(int(status_str, 0) == 0xAA)
                        except ValueError:
                            return Err(Exception(
                                f"Could not parse received status: {status_str}"
                            ))
                    case err:
                        return err
            case err:
                return err

    def verifypinasm(self) -> Result[bool]:
        """
        Verify the PIN code entered by the user by making use of the assembly routine.

        Returns:
            - response (Result): answer from Arduino (True(0xAA) (successful
              attack) or False(0x55) (no double fault)).
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("W"):
                    case Ok(status_str):
                        status_str = status_str.rstrip()
                        try:
                            return Ok(int(status_str, 0) == 0xAA)
                        except ValueError:
                            return Err(Exception(
                                f"Could not parse received status: {status_str}"
                            ))
                    case err:
                        return err
            case err:
                return err

    def verifypinhard(self) -> Result[bool]:
        """
        Verify the PIN code entered by the user by making use of the hardened
        routine.

        Returns:
            - response (Result): answer from Arduino (True(0xAA) (successful
              attack) or False(0x55) (no double fault)).
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("X"):
                    case Ok(status_str):
                        status_str = status_str.rstrip()
                        try:
                            return Ok(int(status_str, 0) == 0xAA)
                        except ValueError:
                            return Err(Exception(
                                f"Could not parse received status: {status_str}"
                            ))
                    case err:
                        return err
            case err:
                return err

    def verifypinstatus(self) -> Result[tuple[int, bool]]:
        """
        Verify the PIN code status.

        Returns:
            - response (Result): tuple of g_ptc and status (True -> Logged as
              admin, False -> Logged as user)
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write("S"):
                    case Ok(_):
                        # Read two consecutive lines
                        match self._safe_readline(), self._safe_readline():
                            case Ok(line1), Ok(line2):
                                line1 = line1.rstrip()
                                line2 = line2.rstrip()
                                try:
                                    return Ok((
                                        int(line1, 0),
                                        int(line2, 0) == 0xAA
                                    ))
                                except ValueError:
                                    return Err(Exception(
                                        f"Could not parse received status:\n{line1=}\n{line2=}"
                                    ))
                            case Err(err), _:
                                return Err(err)
                            case _, Err(err):
                                return Err(err)
                            case _:
                                raise AssertionError(
                                    "Impossible case in verifypinstatus"
                                )
                    case err:
                        return err
            case err:
                return err

    def set_g_ptc_3(self) -> Result[None]:
        """
        Set the maximum number of tries to enter a correct PIN code to 3.

        Returns:
            - response (string): answer from Arduino.
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("3"):
                    case Ok("OK\r\n"):
                        return Ok(None)
                    case Ok(answer):
                        return Err(Exception(
                            f"Received something else than OK: {answer}"
                        ))
                    case err:
                        return err
            case err:
                return err

    def set_noadmin(self) -> Result[None]:
        """
        Close the admin session.

        Returns:
            - response (string): answer from Arduino.
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_read("C"):
                    case Ok("OK\r\n"):
                        return Ok(None)
                    case Ok(answer):
                        return Err(Exception(
                            f"Received something else than OK: {answer}"
                        ))
                    case err:
                        return err
            case err:
                return err

    def setadminpin(self, pin : int) -> Result[None]:
        """
        Changes the admin PIN code in SRAM. Requires admin authentication first.
        Returns Ok(False) if the communication worked, but admin PIN write acces
        was denied, else Ok(True).

        Args:
            - 4 digits PIN to force set as admin PIN.
        """
        # clamp pin to 4 digits
        pin = min(9999, max(0, pin))

        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline(
                    "N" + str(pin).zfill(4),
                    sleep_before_read_s=0.1
                ):
                    case Ok("OK\r\n"):
                        return Ok(None)
                    case Ok(answer):
                        return Err(Exception(
                            f"Invalid PIN or not admin: {answer}"
                        ))
                    case err:
                        return err
            case err:
                return err

    def duplicate_regs_simu(self) -> Result[bool]:
        """
        Starts the duplicate register simulation function on the target.
        By default, the glitcher is setup to do single glitch attacks. It is
        required to go into double glitch mode by using the corresponding
        method of the glitcher class.

        Returns:
            - response (Result): answer from Arduino (True(0xAA) (successful
              attack) or False(0x55) (no double fault)).
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_readline("O", 0.2):
                    case Ok(line):
                        line = line.rstrip()
                        try:
                            return Ok(int(line, 0) == 0xAA)
                        except ValueError:
                            return Err(Exception(
                                f"Could not parse received status: {line}"
                            ))
                    case err:
                        return err
            case err:
                return err

    def encrypt_aes128(self, plain_text: bytes) -> Result[bytes]:
        """
        Perform AES encryption.

        Args:
            - plain_text (bytes): the plain text to be encrypted (as bytes
              `plain_text` should look like b"0123456789ABCDEF").
        """
        if len(plain_text) != 16:
            return Err(Exception("Plain text must be 16 bytes long"))
        else:
            match self._safe_instrument_reset_input_buffer():
                case Ok(_):
                    cmd: bytes = binascii.unhexlify(hex(ord("E"))[2:].encode())
                    match self._safe_bytes_write_then_bytes_read(
                        cmd + plain_text,
                        nb_bytes=16,
                        sleep_before_read_s=0.1
                    ):
                        case Ok(cipher):
                            if len(cipher) != 16:
                                len_cipher = len(cipher)
                                return Err(Exception(
                                    f"Received cipher was not 16 bytes long, was {len_cipher} bytes long."
                                ))
                            else:
                                return Ok(cipher)
                        case err:
                            return err
                case err:
                    return err

    def key_aes128(self, key: bytes) -> Result[None]:
        """
        AES128 key + derivation of rounding keys.
        Must be authentificated as admin on target Arduino!
        Will be ignored otherwise.

        Args:
            - key (string): the AES128 key.

        Returns:
            - response (Result): answer from Arduino.
        """
        if len(key) != 16:
            return Err(Exception("Key must be 16 characters long"))
        else:
            match self._safe_instrument_reset_input_buffer():
                case Ok(_):
                    cmd: bytes = binascii.unhexlify(hex(ord("K"))[2:].encode())
                    match self._safe_bytes_write_then_read(
                        cmd + key,
                        sleep_before_read_s=0.2
                    ):
                        case Ok("OK\r\n"):
                            return Ok(None)
                        case Ok(answer):
                            return Err(Exception(
                                f"Received something else than OK. Are you admin?: {answer}"
                            ))
                        case err:
                            return err
                case err:
                    return err

    def sbox(self) -> Result[bytes]:
        """
        Asks Arduino to send it's SBOX.
        ("0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, ... , 0xb0, 0x54, 0xbb, 0x16")

        Returns:
            - response (Result): SBOX sent by Arduino
        """
        match self._safe_instrument_reset_input_buffer():
            case Ok(_):
                match self._safe_write_then_bytes_read(
                    "B",
                    nb_bytes=256,
                    sleep_before_read_s=0.1
                ):
                    case Ok(sbox_b):
                        if len(sbox_b) != 16 * 16:
                            return Err(Exception(
                                "Received SBOX was not 16*16 bytes long"
                            ))
                        else:
                            return Ok(sbox_b)
                    case err:
                        return err
            case err:
                return err
